//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\CircularFifo.h
// Date: 2024-02-20
// Author: Wolfgang Buescher (DL4YHF)
// Purpose: Simple, classic, lock-free, thread-safe CIRCULAR FIFO
//          with ONE WRITER (in thread A)
//          and MULTIPLE READERS (in threads B,C,D,..)
//          Used in various firmware projects, but since 2024
//          also in the 'Remote CW Keyer' (initially a windows
//          application), to pass data received from, or sent to,
//          a serial port from one thread to another
//          WITHOUT THE NEED FOR THREAD SYNCHRONISATION.
//---------------------------------------------------------------------------

#ifndef  CIRCULAR_FIFO_INCLUDED
# define CIRCULAR_FIFO_INCLUDED

#include "switches.h"  // project specific compiler switches ("options"),
                       // must be included before anything else !
#include "yhf_type.h"  // classic types like BYTE, WORD, DWORD, BOOL, ..

#ifndef  SWI_CIRCULAR_FIFO_SIZE      // if not defined in "switches.h" ..
# define SWI_CIRCULAR_FIFO_SIZE 256  // .. define this configuration parameter HERE
#endif // ndef SWI_CIRCULAR_FIFO_SIZE

typedef struct t_CircularFifo
{
  int iHeadIndex, iTailIndex, iSize;
  BYTE b[SWI_CIRCULAR_FIFO_SIZE]; // the buffer itself must be at the
             // end of the T_CFIFO struct for reasons explained below.

} T_CFIFO;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// If, inside the same application, a few FIFOs with larger sizes
// are required, wrap the T_CFIFO in a larger struct as follows:
// > typedef struct
// > { T_CFIFO fifo;
// > # define LARGE_FIFO_SIZE 16384
// > BYTE bExtraBytes[ LARGER_FIFO_SIZE - SWI_CIRCULAR_FIFO_SIZE ];
// > } T_LARGE_CFIFO
// T_LARGE_CFIFO large_fifo;
//   ...
// CFIFO_Init( &large_fifo.fifo );
// large_fifo.fifo.iSize = LARGE_FIFO_SIZE;
//   ...
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


//----------------------------------------------------------------------------
// Prototypes (callable from C and C++)
//----------------------------------------------------------------------------

#ifndef CPROT  // For peaceful co-existence of "C" and "C++" ....
# ifdef __cplusplus
#   define CPROT extern "C"
# else
#   define CPROT
# endif  // ! "cplusplus"
#endif  // ndef CPROT


CPROT void CFIFO_Init( T_CFIFO *pFifo );
CPROT int  CFIFO_Write(T_CFIFO *pFifo, BYTE *pbSource, int nBytes );
CPROT int  CFIFO_Read( T_CFIFO *pFifo, BYTE *pbDest, int nBytesMax );
CPROT int  CFIFO_GetNumBytesReadable( T_CFIFO *pFifo );

#endif // ndef CIRCULAR_FIFO_INCLUDED ?


/* EOF < CircularFifo.h > */

