// File:  C:\cbproj\Remote_CW_Keyer\HttpServer.h
// Date:  2024-01-20
// Author:  Wolfgang Buescher (DL4YHF)
// Purpose: Socket-based 'Client and Server for the Remote CW Keyer'.
//          As usual, details only in the implementation (CwNet.c) .

#ifndef  _HTTP_SERVER_H  // prevent multiple header inclusion ..
# define _HTTP_SERVER_H  //  .. and let other modules know we're been included

# ifndef _CW_NET_H       // this tiny HTTP-Server is *used by* CwNet.c ...
#  include "CwNet.h"     // .. see the chicken-and-egg problem, A uses B, B uses A
# endif


//---------------------------------------------------------------------------
// Constants
//---------------------------------------------------------------------------
#define HTTP_METHOD_GET     1
#define HTTP_METHOD_POST    2
#define HTTP_METHOD_PUT     3
#define HTTP_METHOD_DELETE  4
#define HTTP_METHOD_OPTIONS 5
#define HTTP_METHOD_HEAD    6


extern const T_SL_TokenList HttpMethods[]; // tokens like "GET ", "POST ", "PUT "


// HTTP "Result codes", actually array indices into HttpSrv_szStatMsgs[] .
// Borland : "Redefinition von 'HTTP_STATUS_OK' ist nicht identisch".
//           Leider gibt der elende Compiler nicht an, in welchem Header
//           dieses Symbol SCHON VORHER definiert war.
//   Nach langer Suche fand sich eine (selbstverstndlich INKOMPATIBLE)
//   Defintion in C:\CBuilder6\Include\Mcf\afxisapi.h . Was fr ein dmlicher Name !
//   Noch eine (ebenfalls inkompatible) Definition stand in 'wininet.h' .
//   Sollen "winninet", "willinet" und "derfinet" bleiben wo der Pfeffer wchst.
//   Zhneknirschend musste WB die folgenden Definitionen (in HttpSrv.h) umbenennen:
//   Alt: #define HTTP_STATUS_OK 0    Neu: #define HTTP_STATUS__OK 0
#define HTTP_STATUS__OK            200 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__CREATED],     "200 OK"                */
#define HTTP_STATUS__CREATED       201 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__CREATED],     "201 Created"           */
#define HTTP_STATUS__ACCEPTED      202 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__ACCEPTED],    "202 Accepted"          */
#define HTTP_STATUS__NOCONTENT     204 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__NOCONTENT],   "204 No Content"        */
#define HTTP_STATUS__MOVEDPERM     301 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__MOVEDPERM],   "301 Moved Permanently" */
#define HTTP_STATUS__MOVEDTEMP     302 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__MOVEDTEMP],   "302 Moved Temporarily" */
#define HTTP_STATUS__NOTMODIFIED   304 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__NOTMODIFIED], "304 Not Modified"      */
#define HTTP_STATUS__BADREQUEST    400 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__BADREQUEST],  "400 Bad Request"       */
#define HTTP_STATUS__UNAUTHORIZED  401 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__UNAUTHORIZED],"401 Unauthorized"      */
#define HTTP_STATUS__FORBIDDEN     403 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__FORBIDDEN],   "403 Forbidden"         */
#define HTTP_STATUS__NOTFOUND      404 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__NOTFOUND],    "404 Not Found"         */
#define HTTP_STATUS__SERVERERROR   500 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__SERVERERROR], "500 Internal Server Error" */
#define HTTP_STATUS__NOTIMPLEMENTED 501 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__NOTIMPLEMENTED],"501 Not Implemented" */
#define HTTP_STATUS__BADGATEWAY    502 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__BADGATEWAY],  "502 Bad Gateway"       */
#define HTTP_STATUS__UNAVAILABLE   503 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__UNAVAILABLE], "503 Service Unavailable" */
#define HTTP_STATUS__SRV_TIMEOUT   504 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__SRV_TIMEOUT], "504 Server or Gateway Timeout" */
#define HTTP_STATUS__SWITCHING_PROTOCOLS 101 /* actually "HTTP/1.1 101" , in HttpSrv_szStatMsgs "101 Switching Protocols" */
#define HTTP_STATUS__INSUFF_STORAGE 506 /* -> in HttpSrv_szStatMsgs[HTTP_STATUS__INSUFF_STORAGE],"507 Insufficient Storage" */
#define HTTP_STATUS__LONG_POLL      600 /* * NOCH KEINE RESPONSE SENDEN ... warten auf 'Long Poll', noch keine sendefhige Response fertig */
#define HTTP_STATUS__SUSPICIOUS_REQ 601 /* "wenn Du noch einmal Bldsinn ausprobierst, fliegst Du fr immer 'raus" */
#define HTTP_STATUS__BLOCK_THIS_BASTARD  602 /* .. and immediately close the connection, without leaking out any info */
        // (the "bastard" will be added to our blacklist, see HttpSrv_ParseRequest() )
#define HTTP_STATUS__NOT_MY_BUSINESS 603  // this value may be returned by
        // specialized server modules if THEY don't want to process the request,
        // but let the default HTTP server's file-request-handler
        //   (e.g. in SpecHttpSrv.cpp : OnHttpFileRequestOrCompletion() )
        // "serve" an ordinary file, loaded from the 'server pages' folder.
extern const T_SL_TokenList HttpStatusCodes[];


typedef enum  // Tokens for keywords in HTTP requests, mostly in the HEADER lines:
{  C_HttpSrv_Kwd_Unknown = 0,
   C_HttpSrv_Kwd_Host,
   C_HttpSrv_Kwd_User_Agent,
   C_HttpSrv_Kwd_Accept,
   C_HttpSrv_Kwd_Accept_Language,
   C_HttpSrv_Kwd_Accept_Encoding,
   C_HttpSrv_Kwd_DNT,
   C_HttpSrv_Kwd_Connection,
   C_HttpSrv_Kwd_Referer,
   C_HttpSrv_Kwd_Content_Type,
   C_HttpSrv_Kwd_Content_Length,
   C_HttpSrv_Kwd_Content_Disposition,
   C_HttpSrv_Kwd_Access_Control_Allow_Origin, // added 2016, important for CORS
   C_HttpSrv_Kwd_Upgrade, // added 2019, important for WebSockets
   C_HttpSrv_Kwd_future
} T_HttpSrv_KeywordToken;

// Possible values parsed from header-line: "Content-Type: ", stored in variables named 'nContentType' .
// Used in string table HttpSrv_ContentTypes[], access via HttpSrv_GetTokenFromTable() / HttpSrv_GetStringFromTable() .
// Later also used as BITWISE COMBINATION for the "Accept"-field in an HTTP request header .
#define T_HttpSrv_ContentType  unsigned long
#define C_HttpSrv_ContentType_Unknown      0
#define C_HttpSrv_ContentType_All          0x00FFFFFF
#define C_HttpSrv_ContentType_Text         0x00000001
#define C_HttpSrv_ContentType_Text_Plain   0x00000002 // -> "Content-Type:text/plain" (also used to POST FORMS!)
#define C_HttpSrv_ContentType_Text_HTML_ISO_8859_1 0x00000004 // -> "Content-Type:text/html; charset=ISO-8859-1"
#define C_HttpSrv_ContentType_Text_HTML    0x00000008 // "text/html" without a 'charset'-specification
#define C_HttpSrv_ContentType_Image        0x00000010
#define C_HttpSrv_ContentType_Image_BMP    0x00000020 // "image/bmp", used by an embedded web server to retrieve the framebuffer
#define C_HttpSrv_ContentType_Image_X_Icon 0x00000040 // "image/x-icon", typically used by a web server's "favicon.ico"
#define C_HttpSrv_ContentType_Video        0x00000080
#define C_HttpSrv_ContentType_Audio        0x00000100
#define C_HttpSrv_ContentType_Audio_X_Wav  0x00000200  // "audio/x-wav"
#define C_HttpSrv_ContentType_Application              0x00000800
#define C_HttpSrv_ContentType_Application_URLEncoded   0x00001000
#define C_HttpSrv_ContentType_Application_JSON         0x00002000 // "application/json", not "text/json" ..
#define C_HttpSrv_ContentType_Application_Octet_Stream 0x00004000 // "application/octet-stream"
#define C_HttpSrv_ContentType_Application_XHTML        0x00008000 // "application/xhtml" (often used by browsers)
#define C_HttpSrv_ContentType_Multipart_FormData 0x00010000
#define C_HttpSrv_ContentType_Multipart_Other    0x00020000 /* not supported ! */
#define C_HttpSrv_ContentType_Message      0x00040000
#define C_HttpSrv_ContentType_XML          0x00100000
#define C_HttpSrv_ContentType_XML_UTF8     0x00200000 // -> "Content-Type:text/xml; charset=utf-8"  (used when returning anything.xml)

// Bitwise combineable flags parsed from header-line "Connection: ",
//  used in string-table HttpSrv_ConnectionTypes[], and for dwConnectionPersistance :
#define C_HttpSrv_ConnectionPersistance_Unknown    0
#define C_HttpSrv_ConnectionPersistance_KeepAlive  1
#define C_HttpSrv_ConnectionPersistance_Close      2
#define C_HttpSrv_ConnectionPersistance_Upgrade    4 /* inevitable for "WebSockets" */

// bitwise combineable flags for keywords in the header-line "Upgrade:" -> dwUpgradeRequested :
#define C_HttpSrv_DontUpgrade       0
#define C_HttpSrv_Upgrade_WebSocket 1
#define C_HttpSrv_Upgrade_TLS       2
#define C_HttpSrv_Upgrade_HTTP2     4

// Bitwise combineable flags parsed from header-line "Disposition-Type" / HttpSrv_DispositionTypes[] :
#define C_HttpSrv_DispositionType_Unknown          0
#define C_HttpSrv_DispositionType_FormData         1
#define C_HttpSrv_DispositionType_Attachment       2



//---------------------------------------------------------------------------
// Data Types (in addition to those in CwNet.h) ..
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// Global variables  (?)
//---------------------------------------------------------------------------

  // (none so far. Most "per-client" data are in T_CwNetClient )


//---------------------------------------------------------------------------
// Function Prototypes
//---------------------------------------------------------------------------

#ifndef CPROT   // for peaceful co-existence of C and C++ ...
#ifdef __cplusplus
 #define CPROT extern "C"
#else
 #define CPROT
#endif  // not "cplusplus" ?
#endif // ndef CPROT ?

CPROT void HttpSrv_OnReceive( T_CwNet *pCwNet, // [in] single SERVER INSTANCE
              T_CwNetClient *pClient,          // [in] instance data for a CLIENT
            BYTE *pbRcvdData, int nBytesRcvd); // [in] received stream segment



#endif // ndef  _HTTP_SERVER_H ?


