//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\Utilities.h
//   (stuff collected from other projects, e.g. DL4YHF's Spectrum Lab)
//---------------------------------------------------------------------------

#ifndef  _YHF_UTILITIES_H
# define _YHF_UTILITIES_H      // prevent multiple inclusion



//-------------- Old-fashioned global variables -----------------------------
//   (use with discretion; don't wrap each and everything into C++ classes)
// Initialized ONLY to prevent chaos if the application
// "forgets" to call UTL_Init() before any other function of this module.
extern int UTL_iDesktopXMin;    // typically something like ZERO
extern int UTL_iDesktopXMax;    // typically something like 1920
extern int UTL_iDesktopYMin;    // typically something like ZERO
extern int UTL_iDesktopYMax;    // typically something like 1080
extern int UTL_iWindowsVersion; // keep it simple: 0=95/98, 1=XP, 2="Vista" (uurgh..),
                                //  7=WIN7, 8=WIN8, 10=WIN10, .. ?
extern int UTL_iAppInstance;   // Application instance index. Possible values:
             // -1 = "Shame on you, haven't called UTL_Init() yet",
             //  0 = "I am the FIRST instance",
             //  1 = "I am the SECOND instance", etc ..
#  define UTL_MAX_APP_INSTANCES 40 // limits UTL_iAppInstance,
             //  and the number of MUTEXES used to count this, in UTL_Init() .

extern char UTL_sz255LastError[256]; // only for debugging .. contains the last error
             // in human readable form after e.g. complex functions like
             // UTL_EnumerateComPorts() experienced unexpected trouble
             // with OS function calls [for example, being unable to extract
             // a COM port's FRIENDLY NAME, which occasionally happened,
             // depending on the Windows version and who-knows-what-else].
extern BOOL UTL_fRunLogOpened; // TRUE when a 'run-log' shall be written
extern char UTL_sz255LogFile[256]; // name optional name of LOGFILE (""=don't TRY TO log)




//-------------- Functions  --------------------------------------------------

#ifndef CPROT  // For peaceful co-existence of "C" and "C++" ....
# ifdef __cplusplus
#   define CPROT extern "C"
# else
#   define CPROT
# endif  // ! "cplusplus"
#endif  // ndef CPROT

CPROT void UTL_Init(void);  // Initializes the module. Call 'as early as you can'.
CPROT void UTL_Exit(void);  // De-initializes the module. Call 'as late as you can'.
CPROT void UTL_LimitInteger( int *pI, int iMin, int iMax );
CPROT DWORD UTL_CRC32( DWORD crc/*~seed*/, BYTE *data, int len );


CPROT double UTL_GetCurrentUnixDateAndTime(void);
CPROT void UTL_FormatDateAndTime( char *format, double dblUnixDateTime, char *dest);
CPROT void UTL_SplitUnixDateTimeToYMDhms( double dblUnixDateTime,
         int *piYear, int *piMonth, int *piDay,
         int *piHour, int *piMinute, double *pdblSecond );


CPROT void UTL_LastWindowsErrorCodeToString( DWORD dwWindowsErrorCode, char *pszDest, int iMaxDestLen );
CPROT void UTL_OpenRunLogFile( char *pszLogFileName );
CPROT BOOL UTL_WriteRunLogEntry( char * pszFormatString, ... );

CPROT void UTL_ClipPointToDesktop( int *piX, int *piY); // Clips a pixel coordinate into the "total" desktop areavoid UTL_ClipXYWHToDesktop( int *piX, int *piY, int *piWidth, int *piHeight ); // Clips an entire window..

CPROT BOOL UTL_EnumerateComPorts(int *piEnumerator, char *pszComPortName, int iMaxLen);
CPROT int  UTL_ParseNumberFromComPortName(char *pszComPortName);
CPROT int  UTL_CompareComPortNames( char *pszComPortName1, char *pszComPortName2);


CPROT BOOL UTL_GetPathToMyDataFiles( char *pszDest, int iMaxLen );


//--------------------------------------------------------------------------
// Helpers to deal with old-fashioned INI-files (for 'small config files')
//--------------------------------------------------------------------------

CPROT const char *UTL_IntToStr(long i32);
CPROT long UTL_ReadIntFromIniFile(char *pszIniFileName, char *pszSection, char *pszKeyName, long i32Default);
CPROT void UTL_WriteIntToIniFile(char *pszIniFileName, char *pszSection, char *pszKeyName, long i32Value);
CPROT void UTL_ReadStringFromIniFile(char *pszIniFileName, char *pszSection, char *pszKeyName,
                            char *pszDefault,  char *pszResult, int iMaxLen );
CPROT void UTL_WriteStringToIniFile(char *pszIniFileName, char *pszSection, char *pszKeyName, char *pszValue );

//--------------------------------------------------------------------------
// Helpers for Microsoft's "RichText" edit control (without VCL, just Win32)
//--------------------------------------------------------------------------
CPROT void RichEdit_SetSelBgColor( HWND hWndEditor, DWORD dwColor );
CPROT void RichEdit_SetSelColors( HWND hWndEditor, DWORD dwTextColor, DWORD dwBkgndColor );


//---------------------------------------------------------------------------
CPROT void ShowError( int iErrorClass, char * pszFormat, ... );
  // ,---------------------------'
  // '--> ERROR_CLASS_FATAL, ERROR_CLASS_ERROR, ERROR_CLASS_WARNING, ERROR_CLASS_INFO .
  //
  // Not implemented in Utilities.c but "somewhere in the application"
  // (because only "the application" knows where, and how, to output
  //  those messages in a relatively thread-safe way, ideally 'colourized')
# define ERROR_CLASS_FATAL     4
# define ERROR_CLASS_ERROR     3
# define ERROR_CLASS_WARNING   2
# define ERROR_CLASS_INFO      1
# define ERROR_CLASS_ALL       0 /* show ANYTHING */
  // Since 2024-01, the upper bits in the first argument for ShowError()
  // specify whether to emit the message into the "debug run log" or not,
  // if ShowError() shall prefix the string with a timestamp, etc:
# define SHOW_ERROR_IN_RUN_LOG 0x8000 // -> Let ShowError() pass the message on to UTL_WriteRunLogEntry()
# define SHOW_ERROR_TIMESTAMP  0x4000 // -> Let ShowError() insert a TIMESTAMP (time of day?) before the message





#endif // ndef _YHF_UTILITIES_H ?

/* EOF <Utilities.h> */

