// File: C:\cbproj\Remote_CW_Keyer\dsound_wrapper.c
// Purpose: Wrapper between non-MFC applications and "Direct Sound" .
// Based on dsound_wrapper.c,v 1.1.1.1 2002/01/22 00:52:45,
//          "Simplified DirectSound interface",
//          by Phil Burk & Robert Marsanyi .
//   (in 2023, no traces of the above file were found in the WWW anymore)
//    Original description quoted below - DON'T MODIFY -,
//    plus various additions by DL4YHF who tried to understand
//    how this is supposed to work (any why it didn't work)
//    *further below* .
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// PortAudio Portable Real-Time Audio Library
// For more information see: http://www.softsynth.com/portaudio/
// DirectSound Implementation
// Copyright (c) 1999-2000 Phil Burk & Robert Marsanyi
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// Any person wishing to distribute modifications to the Software is
// requested to send the modifications to the original developer so that
// they can be incorporated into the canonical version.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
// ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//
// Notes (by DL4YHF, 2023):
//   * As already expected, when really INVOKING any of the functions in
//     'dsound_wrapper.c', Borland C++ Builder 6 complained :
//       > Unresolved external 'DirectSoundEnumerateA referenced from .. bla..'
//       > Unresolved external 'DirectSoundCreate referenced from .. blubb..'
//       > Unresolved external 'DirectSoundCaptureCreate referenced from .. '
//     Microsoft (formerly MSDN, not "learn.microsoft.com") suggested
//     to use 'Link Library: Dsound.lib'. A-ha. Where's that ?
//     Fortunately, even the stoneage Borland C++Builder 6 was shipped with
//     this library, so added C:\CBuilder6\Lib\Psdk\dsound.lib to the project,
//     hoping THAT LIBRARY uses a calling convention compatible with BCB .
//
//   * Even though many (including Microsoft) say 'DirectSound is a dead horse',
//     some say it's better than the stoneage WMME 'Wave Audio API',
//     as long as low audio latency is concerned. For the planned use
//     (generation of a low-latency 'Morse sidetone output' for an electronic
//      keyer), low audio latency was the primary goal, WITHOUT having to dig
//      into the hyped "AudioGraph settings or the WASAPI IAudioClient3 interface",
//      even a dead horse may be better than a dead dinosaur. Time will tell.
//      Audio latency test results are possibly documented in
//      C:\cbproj\Remote_CW_Keyer\KeyerThread.c .
//

#include "switches.h" // project specific 'compilation switches' like
                      // SWI_USE_DSOUND_WRAPPER, SWI_HARDCORE_DEBUGGING, etc.

#ifdef __BORLANDC__
// prevent that bugger from bugging us with 'initialized data in header, cannot pre-compile..." :
# pragma hdrstop
#endif


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define INITGUID     // Needed to build IID_IDirectSoundNotify. See objbase.h for info.
#include <objbase.h> // e.g. in c:\cbuilder6\include\objbase.h (a windows thing)
#include <unknwn.h>  // e.g. in c:\cbuilder6\include\unknwn.h  (also a windows thing)
#include "dsound_wrapper.h"

// ex: #include "pa_trace.h"
#define AddTraceMessage(msg,data) /* no 'trace' available */

//----------------------------------------------------------------------------
// Global variables for hardcore debugging / post-mortem crash analysis.
//        Useful when the debugger's call stack shows nothing but garbage,
//        e.g. after an exception at the infamous address 0xFEEEFEEE .
//        ( C++Builder's debugger could still inspect SIMPLE GLOBAL
//          VARIABLES after most kinds of exceptions, but of course nothing
//          stack-based because the CPU- or task-stack was usually trashed.)
//----------------------------------------------------------------------------
#if(SWI_HARDCORE_DEBUGGING) // (1) = hardcore-debugging, (0)=normal compilation
 int DSW_iLastSourceLine = 0; // WATCH THIS after crashing with e.g. "0xFEEEFEEE"  ...
# define HERE_I_AM__DSW()  DSW_iLastSourceLine=__LINE__
  // (to also WATCH the instance of DSoundWrapper that most likely crashed,
  //  in Borland C++Builder, add a 'watch' for MyDirectSound ,
  //  especially examine ..
  //     MyDirectSound.fInputOpened   MyDirectSound.pDirectSoundIn
  //     MyDirectSound.pInputBuffer   MyDirectSound.dwInputFifoSizeInBytes
  //     MyDirectSound.uReadOffset .
  // See also: Complete list of other XYZ_iLastSourceLine varibles to watch
  //     in C:\cbproj\Remote_CW_Keyer\Keyer_Main.cpp, near GUI_iLastSourceLine)
#else
# define HERE_I_AM__DSW()
#endif // SWI_HARDCORE_DEBUGGING ?


// internal function prototypes (may be decorated with 'static'...)
void DSW_OnOutputError( DSoundWrapper *dsw, char *pszInfo, HRESULT hr );
void DSW_SetFirstOutputError( DSoundWrapper *dsw, char *pszInfo );
void DSW_OnInputError( DSoundWrapper *dsw, char *pszInfo, HRESULT hr );
void DSW_SetFirstInputError( DSoundWrapper *dsw, char *pszInfo );


//---------------------------------------------------------------------------
// Common functions used for audio-in- as well as -output:
//---------------------------------------------------------------------------

static LONGLONG DSW_i64HighResTimerOffset = 0; // kludge to let our 'internal timestamps' start at ZERO

//---------------------------------------------------------------------------
double DSW_ReadHighResTimestamp_s( void ) // <- the "_s" suffix is the physical unit
  // Actually just a 'thin abstraction layer' to the
  //  "QueryPerformanceCounter"-thingy in Windows, with a conversion into
  //  the one-and-only SI unit for time (no machine-depending 'tick frequency').
{
  LONGLONG i64, i64Freq;
  double dblSeconds;

  QueryPerformanceCounter( (LARGE_INTEGER*)&i64 );
  QueryPerformanceFrequency( (LARGE_INTEGER*)&i64Freq ); // will the FREQUENCY change when speed-stepping ? (hope not..)
  if( DSW_i64HighResTimerOffset == 0 )
   {  DSW_i64HighResTimerOffset = i64;
   }
  if( i64Freq > 0 )
   { // ex: i64 = (i64 * 1000000LL) / i64Freq; // even a 64-bit product may be insufficient..
     // thus use floating point to scale from timer ticks into microseconds:
     dblSeconds = (double)(i64 - DSW_i64HighResTimerOffset) / (double)i64Freq;
     //  ,----------------|_______________________________|
     //  '--> difference in timer ticks (tick frequency MAY be many Megahertz..
     //       thus do the SCALING into seconds with 64-bit *double* )
   }
  else  // there doesn't seem to be a 'performance counter / timer' on this system !
   { dblSeconds = 0; // can this happen ? guess not, but Microsoft said "be prepared for the worst"..
   }
  return dblSeconds;
} // end DSW_ReadHighResTimestamp_s()

//---------------------------------------------------------------------------
int CALLBACK DSW_EnumOutputsProc(  // *internal* - no API function !
        LPGUID lpGUID,      // [in] > Address of the GUID that identifies the
                            //      > DirectSound driver being enumerated.
                            //      > This value can be passed to DirectSoundCreate() ...
        LPCTSTR lpszDesc,   // [in] > Address of a null-terminated string
                            //      > that provides a textual description
                            //      > of the DirectSound OUTPUT device.
        LPCTSTR lpszDrvName,// [in] > Address of a null-terminated string that
                            //      > specifies the module name of the DirectSound
                            //      > OUTPUT driver corresponding to this device.
        LPVOID lpContext )  // [in] our application-defined "instance data"
{ DSoundWrapper *dsw = (DSoundWrapper*)lpContext;
  // > Remarks
  // > The application can save the strings passed in the lpcstrDescription
  // > and lpcstrModule parameters by copying them to memory allocated from
  // > the heap. The memory used to pass the strings to this callback function
  // > is valid only while this callback function is running.
  if( (dsw!=NULL) && (lpszDesc!=NULL) )
   {
     if( dsw->m_nOutputDevices < DSW_MAX_ENUMERATED_DEVICES )
      { strncpy( dsw->m_sOutputDevices[dsw->m_nOutputDevices].sz255DevName,
                 lpszDesc, 255 ); // seen here: "Primary Sound Driver",
                 // but the Primary Sound Driver's GUID was a NULL pointer.
                 // Don't try to COPY the "GUID" from such a pointer ... :
        if( lpGUID == NULL ) // this guy doesn't have a guid .. guess he is the "Primary Sound Driver" (output; "renderer")
         { memset( &dsw->m_sOutputDevices[dsw->m_nOutputDevices].guid,
                   0, sizeof( GUID ) ); // "all zeroes" in here will later be replaced by a NULL pointer
         }
        else // not sure if the "GUID pointer" remains valid until we need it,
         {   // so BLOCK-COPY this 128-bit "globally unique identifier" :
           dsw->m_sOutputDevices[dsw->m_nOutputDevices].guid = *lpGUID;
           // The first (output) devices that DID have a "GUID" were :
           //  * "MAGIC LCD (Intel(R) Display-Audio)"
           //  * "Lautsprecher (USB Audio CODEC)"
           //  * "Lautsprecher (Realtek High Definition Audio)"
           //  * "Lautsprecher/Kopfhrer (VB-Audio Virtual Audio Cable)"
         }
        ++dsw->m_nOutputDevices;
        // > Return Values (of the DSEnumCallback) :
        // > Returns TRUE to continue enumerating drivers, or FALSE to stop.
        return ( dsw->m_nOutputDevices < DSW_MAX_ENUMERATED_DEVICES );
      }
   }
  // Arrived here ? Too many devices, or we're unhappy with what was passed in !
  return FALSE; // stop enumerating more DirectSound OUTPUT devices !
} // end DSW_EnumOutputsProc()

//---------------------------------------------------------------------------
int CALLBACK DSW_EnumInputsProc(  // *internal* - no API function !
        LPGUID lpGUID,      // .. same meaning as in DSW_EnumOutputsProce() ..
        LPCTSTR lpszDesc, LPCTSTR lpszDrvName, LPVOID lpContext )
{ DSoundWrapper *dsw = (DSoundWrapper*)lpContext;
  if( (dsw!=NULL) && (lpszDesc!=NULL) )
   {
     if( dsw->m_nInputDevices < DSW_MAX_ENUMERATED_DEVICES )
      { strncpy( dsw->m_sInputDevices[dsw->m_nInputDevices].sz255DevName,
                 lpszDesc, 255 );
        if( lpGUID == NULL ) // this guy doesn't have a guid .. guess he is the "Primary Sound Capture Driver"..
         { memset( &dsw->m_sInputDevices[dsw->m_nInputDevices].guid,
                   0, sizeof( GUID ) ); // "all zeroes" in here will later be replaced by a NULL pointer
         }
        else // this audio "capturing" (input) decive has a guid, so COPY it (128 bits, not just the address):
         { dsw->m_sInputDevices[dsw->m_nInputDevices].guid = *lpGUID;
           // The first (input-)"capturing" devices that DID have a "GUID" were :
           //  * "Onboard-Microphone (Realtek Hig" [same stupid truncation as in the WMM "wave audio" device)
           //  * "CABLE Output (VB-Audio Virtual "
           //  * "Mikrofon (USB Audio CODEC)"  [believe it or not, THIS was an Icom IC-7300]
         }
        ++dsw->m_nInputDevices;
        // > Return Values (of the DSEnumCallback) :
        // > Returns TRUE to continue enumerating drivers, or FALSE to stop.
        return ( dsw->m_nInputDevices < DSW_MAX_ENUMERATED_DEVICES );
      }
   }
  return FALSE; // stop enumerating more DirectSound INPUT devices !
} // end DSW_EnumInputsProc()


//---------------------------------------------------------------------------
BOOL DSW_EnumerateDevices( DSoundWrapper *dsw )                       // API
  // Tries to enumerate all DirectSound capable audio in- and output devices.
  // Originally a thin wrapper for Microsoft's "DirectSoundEnumerate()" .
  // To also enumerate INPUT devices, "DirectSoundCaptureEnumerate()" is used.
  // After this, an application can list all 'DirectSound' capable devices
  //    in e.g. combo lists for the IN- and OUTPUT device .
  //
  // Returns TRUE on success, FALSE otherwise.
  //  [out] dsw->m_nOutputDevices : number of OUTPUT devices (for "rendering"),
  //        dsw->m_nInputDevices  : number of INPUT  devices (for "capturing"),
  //        dsw->m_sOutputDevices[] : names and GUIDs of OUTPUT devices,
  //        dsw->m_sInputDevices[]  : names and GUIDs of INPUT devices.
  //
  // We avoid the name "capture", because from an application's point of view,
  // those are INPUT devices for almost anything (radio receivers, SDRs,
  // microphones, "line inputs", mosquito detectors) - not only for "capturing".
{
  dsw->m_nOutputDevices = dsw->m_nInputDevices = 0;
  DirectSoundEnumerate( DSW_EnumOutputsProc, (void*)dsw );
  DirectSoundCaptureEnumerate( DSW_EnumInputsProc, (void*)dsw );
  return ( dsw->m_nOutputDevices + dsw->m_nInputDevices ) > 0;

} // end DSW_EnumerateDevices()

//---------------------------------------------------------------------------
void DSW_Term( DSoundWrapper *dsw )                                   // API
{
  // Cleanup the sound buffers, for both in- and output
  DSW_CloseInput( dsw );   // <- doesn't do harm if NOT opened at all ..
  DSW_CloseOutput( dsw );
} // end DSW_Term()


//---------------------------------------------------------------------------
BOOL DSW_Init( DSoundWrapper *dsw )                                   // API
{
  memset( dsw, 0, sizeof(DSoundWrapper) );
  dsw->dblLastOutputTime_s = dsw->dblLastInputTime_s = DSW_ReadHighResTimestamp_s();
  return TRUE;
}

//---------------------------------------------------------------------------
char *DSW_ErrorCodeToString( HRESULT hr )
{
  switch( hr )
   { case DS_OK : return "";

     case DS_NO_VIRTUALIZATION : // The function completed successfully, but we had to substitute the 3D algorithm
          return "no virtualization";

     case DSERR_ALLOCATED:    // The call failed because resources (such as a priority level)
          return "allocated"; // were already being used by another caller

     case DSERR_CONTROLUNAVAIL: // The control (vol,pan,etc.) requested by the caller is not available
          return "control unavailable";

     case DSERR_INVALIDPARAM: // An invalid parameter was passed to the returning function
          return "invalid parameter";

     case DSERR_INVALIDCALL:  // This call is not valid for the current state of this object
          return "invalid call";

     case DSERR_GENERIC:      // An undetermined error occured inside the DirectSound subsystem
          return "generic error";

     case DSERR_PRIOLEVELNEEDED: // The caller does not have the priority level required for the function to succeed
          return "insufficient priority level";

     case DSERR_OUTOFMEMORY:  // Not enough free memory is available to complete the operation
          return "out of memory";

     case DSERR_BADFORMAT:    // The specified WAVE format is not supported
          return "bad wave format";

     case DSERR_UNSUPPORTED:  // The function called is not supported at this time
          return "unsupported";

     case DSERR_NODRIVER:     // No sound driver is available for use
          return "no driver available";

     case DSERR_ALREADYINITIALIZED: // This object is already initialized
          return "already initialized";

     case DSERR_NOAGGREGATION: // This object does not support aggregation
          return "no aggregation";

     case DSERR_BUFFERLOST:    // The buffer memory has been lost, and must be restored
          return "buffer lost";

     case DSERR_OTHERAPPHASPRIO: // Another app has a higher priority level, preventing this call from succeeding
          return "other app has priority";

     case DSERR_UNINITIALIZED: // This object has not been initialized
          return "uninitialized";

     case DSERR_NOINTERFACE:   // The requested COM interface is not available
          return "no COM interface";

     case DSERR_ACCESSDENIED:  // Access is denied
          return "access denied";

     default:  return "???";   // looks like our copy of "dsound.h" is outdated :)
   } // end switch
} // end DSW_ErrorCodeToString()

//---------------------------------------------------------------------------
// Audio OUTPUT ("Rendering Audio" a la DirectSound) ... usage summary :
// 1.) DSW_OpenOutputDevice() with the DESCRIPTIVE NAME (not a "GUID") .
// 2.) Set sample rate, number of channels, and buffer size (thus latency)
//        per DSW_InitOutputBuffer(). In Microsoft geek speak, that's
//        IDirectSound_CreateSoundBuffer() for the rarely-used "primary",
//        and (more important) for the "secondary" buffer.
//        DSW_InitOutputBuffer() fills the buffer with 'silence'
// 4.) Start "streaming out" per DSW_StartOutput() [ -> IDirectSoundBuffer_Play(), Loop ].
//        From this moment on, periodically 'feed' the buffer,
//        e.g. by calling DSW_WriteBlock() [ -> IDirectSoundBuffer_Lock(), Write,
//                                              IDirectSoundBuffer_Unlock() ]
// Quoted from an old "Microsoft DirectX 9.0" SDK:
//  > Using Streaming Buffers
//  > A streaming buffer plays a long sound that cannot all fit into the buffer
//  > at once. As the buffer plays, old data is periodically replaced with new data.
//  > To play a streaming buffer, call the IDirectSoundBuffer8::Play() method,
//  > specifying DSBPLAY_LOOPING in the dwFlags parameter.
//  > To halt playback, call the IDirectSoundBuffer8::Stop() method.
//  > This method stops the buffer immediately, so you need to be sure that
//  > all data has been played. This can be done by polling the play position
//  > or by setting a notification position.
//  > Streaming into a buffer requires the following steps:
//  > 1. Ascertain whether the buffer is ready to receive new data.
//  >    This can be done either by polling the play cursor
//  >    or by waiting for a notification.
//  > 2. Lock a portion of the buffer by using IDirectSoundBuffer8::Lock().
//  >    This method returns one or two addresses where data can now be written.
//  > 3. Write the audio data to the returned address or addresses
//  >    by using a standard memory-copy routine.
//  > 4. Unlock the buffer using IDirectSoundBuffer8::Unlock() .
//  > The reason IDirectSoundBuffer8::Lock might return two addresses is that
//  > you can lock any number of bytes, up to the size of the buffer,
//  > regardless of the start point. If necessary, the locked portion
//  > wraps around to the beginning of the buffer. If it does, you have to
//  > perform two separate memory copies.
//  > For example, say you lock 30,000 bytes beginning at offset 20,000
//  > in a 40,000-byte buffer. When you call Lock in this case,
//  > it returns four values:
//  >   * The memory address of offset 20,000.
//  >   * The number of bytes locked from that point to the end of the buffer;
//  >     that is, 20,000 bytes. You write this number of bytes to the first address.
//  >   * The memory address of offset 0.
//  >   * The number of bytes locked from that point; that is, 10,000 bytes.
//  >     You write this number of bytes to the second address.
//  > If no wraparound is necessary, the last two values are NULL and 0 respectively.
//  > Although it's possible to lock the entire buffer, you must not do so
//  > while it is playing. Normally you refresh only a fraction of the buffer
//  > each time. For example, you might lock and write to the first quarter
//  > of the buffer as soon as the play cursor reaches the second quarter,
//  > and so on. You must never write to the part of the buffer that lies
//  > between the play cursor and the write cursor.
//     (On first glance,
//      the "write cursor" looks like a circular buffer's "FIFO HEAD INDEX",
//      the "play cursor" looks like a circular buffer's "FIFO TAIL INDEX".
//      but that's NOT the case - so read on):
//  > DirectSound maintains two pointers into the buffer: the play cursor
//  > and the write cursor. These positions are byte offsets into the buffer,
//  > not absolute memory addresses.  (...)
//  > The write cursor is the point after which it is safe to write data into the buffer.
//  > The block between the play cursor and the write cursor is already committed
//  > to be played, and cannot be changed safely.
//  >
//  > You might visualize the buffer as a clock face, with data written to it
//  > in a clockwise direction. The play cursor and the write cursor are like
//  > two hands sweeping around the face at the same speed, the write cursor
//  > always keeping a little ahead of the play cursor. If the play cursor
//  > points to the 1 and the write cursor points to the 2, it is only safe
//  > to write data after the 2. Data between the 1 and the 2 may already
//  > have been queued for playback by DirectSound and should not be touched.
//  >
//  > The write cursor moves with the play cursor, not with data written to the buffer.
//  > If you're streaming data, you are responsible for maintaining
//  > your own pointer into the buffer to indicate where the next block of data
//  > should be written.
//  >
//  > An application can retrieve the play and write cursors by calling
//  > the IDirectSoundBuffer8::GetCurrentPosition() method.
//  > The IDirectSoundBuffer8::SetCurrentPosition() method lets you move
//  > the play cursor. Applications do not control the position of the write cursor.
//  >
//  > To ensure that the play cursor is reported as accurately as possible,
//  > always specify the DSBCAPS_GETCURRENTPOSITION2 flag when creating
//  > a secondary buffer. For more information, see DSBUFFERDESC.
//  >
//  > Play Buffer Notification
//  >
//  > DirectSound can notify your application when the play cursor reaches
//  > certain points in a buffer, or when the buffer stops. This can be useful
//  > in two common situations:
//  >  * The application must take some action, such as playing another buffer, 7
//  >    when a sound has finished playing.
//  >  * The application is streaming data to the buffer, and needs to know
//  >    when the play cursor reaches predefined points so that data
//  >    that has been played can be replaced by new data.
//  > Using the IDirectSoundNotify8::SetNotificationPositions() method, you can
//  > set any number of points within the buffer where events are to be signaled.
//  > You cannot do this while the buffer is playing. (...)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
BOOL DSW_OpenOutputDevice(
           DSoundWrapper *dsw,
           char *sz255DevName ) // [in] descriptive audio *DEVICE NAME* (not a stuid 'GUID')
  // Return value : TRUE = ok,   FALSE = failure (e.g. device name not found) .
  //     If the function failed, dsw->sz255LastOutputError may give a clue why.
  //
  // Microsoft explains, about DirectSoundCreate() :
  // > This function creates and initializes an IDirectSound interface.
  // > Parameters:
  // > * lpGuid
  // >      Address of the GUID that identifies the sound device.
  // >      The value of this parameter must be one of the GUIDs
  // >      returned by DirectSoundEnumerate(), or NULL for the
  // >      default device.
  // > * ppDS (2nd argument)
  // >      Address of a pointer to a DirectSound object
  // >      created in response to this function.
  // > * pUnkOuter
  // >      Controlling unknown of the aggregate. Its value must be NULL.
  // >      (no idea what an "aggregate" is, guess a Microsoft 'COM' thing)
  // > Remarks
  // >   The application must call IDirectSound::SetCooperativeLevel()
  // >   immediately after creating a DirectSound object.
  //     (WB: .. with DSSCL_EXCLUSIVE, DSSCL_NORMAL, etc.
  //      In this 'wrapper', that happens in DSW_InitOutputBuffer(). )
{
  int  i;
  GUID *pGUID = NULL;
  HRESULT hr;
  BOOL found_device = FALSE;


  if( sz255DevName != NULL ) // use a specific device (not the "Default"-thingy) ?
   { if( dsw->sz255OutputDeviceName != sz255DevName ) // <- compare ADDRESS, not STRING CONTENT
      { strncpy( dsw->sz255OutputDeviceName, sz255DevName, 255 );
        // '--> dinosaur will crash when copying a string to itself !
      }
     for(i=0;( (i<dsw->m_nOutputDevices) && (i<DSW_MAX_ENUMERATED_DEVICES) ); ++i)
      { if( strcmp(sz255DevName, dsw->m_sOutputDevices[i].sz255DevName) == 0 )
         { if( dsw->m_sOutputDevices[i].guid.Data1 != 0 )
            { pGUID = &dsw->m_sOutputDevices[i].guid;
            }
           found_device = TRUE;
           break;
         }
      }
   }    // end if( sz255DevName != NULL )
  else
   { dsw->sz255OutputDeviceName[0] = '\0';
   }

  if( ! found_device )
   { strcpy( dsw->sz255LastOutputError, "OpenOutput: Device not found" );
     return FALSE;
   }

  // Create the DS object
  hr = DirectSoundCreate( pGUID, &dsw->pDirectSoundOut, NULL );
  if( hr == DS_OK )
   { return TRUE;
   }
  return FALSE;
} // end DSW_OpenOutputDevice()


//---------------------------------------------------------------------------
BOOL DSW_InitOutputBuffer(
           DSoundWrapper *dsw,
           unsigned long nFrameRate, // Here, a "frame" is a "SAMPLE POINT", possibly with MULTIPLE CHANNELS,
                                     // all sampled at the same time (if the hardware permits).
           int nChannels,
           int bytesPerBuffer )
{
  DWORD          dwDataLen;
  DWORD          playCursor;
  HRESULT        hr;
  LPDIRECTSOUNDBUFFER pPrimaryBuffer;
  HWND           hWnd;
  WAVEFORMATEX   wfFormat;
  DSBUFFERDESC   primaryDesc;
  DSBUFFERDESC   secondaryDesc;
  unsigned char* pDSBuffData;
  LARGE_INTEGER  counterFrequency;

  // Sanity check (by DL4YHF) to catch 'application errors' :
  if( dsw->pDirectSoundOut == NULL ) // oops.. output device not opened !
   { DSW_SetFirstOutputError( dsw, "InitBuffer called but Output not open" );
     return FALSE;
   }


  dsw->iOutputSize = bytesPerBuffer;
  dsw->fOutputOpened  = FALSE;
  dsw->nOutputUnderflows = 0;
  dsw->dblSamplesWrittenToOutput = 0;
  dsw->dwBytesPerFrame_out = nChannels * sizeof(short);
  dsw->dblLastOutputTime_s = DSW_ReadHighResTimestamp_s();


  // We were using getForegroundWindow() but sometimes the ForegroundWindow may not be the
  // applications's window. Also if that window is closed before the Buffer is closed
  // then DirectSound can crash. (Thanks for Scott Patterson for reporting this.)
  // So we will use GetDesktopWindow() which was suggested by Miller Puckette.
  // hWnd = GetForegroundWindow();
  hWnd = GetDesktopWindow();

  // Set cooperative level to DSSCL_EXCLUSIVE so that we can get 16 bit output, 44.1 KHz.
  // Exclusive also prevents unexpected sounds from other apps during a performance.
  if ((hr = IDirectSound_SetCooperativeLevel( dsw->pDirectSoundOut,
              hWnd, DSSCL_EXCLUSIVE)) != DS_OK)
   { DSW_OnOutputError( dsw, "SetCooperativeLevel", hr);
     return FALSE;
   }
  // -----------------------------------------------------------------------
  // Create primary buffer and set format just so we can specify our custom format.
  // Otherwise we would be stuck with the default which might be 8 bit or 22050 Hz.
  // Setup the primary buffer description
  ZeroMemory(&primaryDesc, sizeof(DSBUFFERDESC));
  primaryDesc.dwSize        = sizeof(DSBUFFERDESC);
  primaryDesc.dwFlags       = DSBCAPS_PRIMARYBUFFER; // all panning, mixing, etc done by synth
  primaryDesc.dwBufferBytes = 0;
  primaryDesc.lpwfxFormat   = NULL;
  // Create the buffer
  if ((hr = IDirectSound_CreateSoundBuffer( dsw->pDirectSoundOut,
                  &primaryDesc, &pPrimaryBuffer, NULL)) != DS_OK)
   { DSW_OnOutputError( dsw, "CreateSoundBuffer", hr);
     return FALSE;
   }

  // Define the buffer format
  wfFormat.wFormatTag = WAVE_FORMAT_PCM;
  wfFormat.nChannels = nChannels;
  wfFormat.nSamplesPerSec = nFrameRate;
  wfFormat.wBitsPerSample = 8 * sizeof(short);
  wfFormat.nBlockAlign = wfFormat.nChannels * wfFormat.wBitsPerSample / 8;
  wfFormat.nAvgBytesPerSec = wfFormat.nSamplesPerSec * wfFormat.nBlockAlign;
  wfFormat.cbSize = 0;  /* No extended format info. */

  // Set the primary buffer's format
  if((hr = IDirectSoundBuffer_SetFormat( pPrimaryBuffer, &wfFormat)) != DS_OK)
   { DSW_OnOutputError( dsw, "SetFormat", hr);
     return FALSE;
   }

  // ----------------------------------------------------------------------
  // Setup the secondary buffer description
  ZeroMemory(&secondaryDesc, sizeof(DSBUFFERDESC));
  secondaryDesc.dwSize = sizeof(DSBUFFERDESC);
  secondaryDesc.dwFlags =  DSBCAPS_GLOBALFOCUS | DSBCAPS_GETCURRENTPOSITION2;
  secondaryDesc.dwBufferBytes = bytesPerBuffer;
  secondaryDesc.lpwfxFormat = &wfFormat;

  // Create the secondary buffer
  if ((hr = IDirectSound_CreateSoundBuffer( dsw->pDirectSoundOut,
                  &secondaryDesc, &dsw->pOutputBuffer, NULL)) != DS_OK)
   { DSW_OnOutputError( dsw, "CreateSoundBuffer", hr);
     return FALSE;
   }

  // Lock the DS buffer
  if ((hr = IDirectSoundBuffer_Lock( dsw->pOutputBuffer, 0, dsw->iOutputSize, (LPVOID*)&pDSBuffData,
                                           &dwDataLen, NULL, 0, 0)) != DS_OK)
   { DSW_OnOutputError( dsw, "LockBuffer", hr);
     return FALSE;
   }

  // Zero the DS buffer
  ZeroMemory(pDSBuffData, dwDataLen);

  // Unlock the DS buffer .. we'll "Unlock" buffers in other functions, too,
  //                         so here's what "Unlock" means in this context.
  //     From the horse's mouth ("learn.microsoft.com", replaced MSDN) :
  // > The IDirectSoundBuffer's Unlock() method releases a locked sound buffer.
  // >  [in] lpvAudioPtr1 :
  // >       Address of the value retrieved in the lplpvAudioPtr1 parameter
  // >       of the IDirectSoundBuffer::Lock method.
  // >  [in] dwAudioBytes1
  // >       Number of bytes actually written to the lpvAudioPtr1 parameter.
  // >       It should not exceed the number of bytes returned by the IDirectSoundBuffer::Lock method.
  // >  [in] lpvAudioPtr2
  // >       Address of the value retrieved in the lplpvAudioPtr2 parameter
  // >       of the IDirectSoundBuffer::Lock method.
  // >  [in] dwAudioBytes2
  // >       Number of bytes actually written to the lpvAudioPtr2 parameter.
  // >       It should not exceed the number of bytes returned by the
  // >       IDirectSoundBuffer::Lock method.
  // >
  // > Remarks (still about IDirectSoundBuffer::Unlock(Ptr1,Bytes1,Ptr2,Bytes2) ) :
  // > An application must pass both pointers, lpvAudioPtr1 and lpvAudioPtr2,
  // > returned by the IDirectSoundBuffer::Lock method to ensure the correct
  // > pairing of IDirectSoundBuffer::Lock and IDirectSoundBuffer::Unlock.
  // > The second pointer is needed even if 0 bytes were written to the second pointer.
  // > Applications must pass the number of bytes actually written to the two
  // > pointers in the parameters dwAudioBytes1 and dwAudioBytes2.
  // > Make sure the sound buffer does not remain locked for long periods of time.
  // >
  if ((hr = IDirectSoundBuffer_Unlock( dsw->pOutputBuffer, pDSBuffData, dwDataLen, NULL, 0)) != DS_OK)
   { DSW_OnOutputError( dsw, "UnlockBuffer", hr);
     return FALSE;
   }

  if( QueryPerformanceFrequency( &counterFrequency ) )
   {
     int framesInBuffer = bytesPerBuffer / (nChannels * sizeof(short));
     dsw->i64CounterTicksPerOutputBuffer.QuadPart = (counterFrequency.QuadPart * framesInBuffer) / nFrameRate;
     AddTraceMessage("i64CounterTicksPerOutputBuffer = %d\n", dsw->i64CounterTicksPerOutputBuffer.LowPart );
   }
  else
   {
     dsw->i64CounterTicksPerOutputBuffer.QuadPart = 0;
   }

  // Let DSound set the starting write position because if we set it to zero, it looks like the
  // buffer is full to begin with. This causes a long pause before sound starts when using large buffers.
  hr = IDirectSoundBuffer_GetCurrentPosition( dsw->pOutputBuffer, &playCursor, &dsw->dwWriteOffset );
  if( hr != DS_OK )
   { DSW_OnOutputError( dsw, "GetCurrentPosition", hr);
     return FALSE;
   }
  dsw->dblSamplesWrittenToOutput = dsw->dwWriteOffset / dsw->dwBytesPerFrame_out;
  /* printf("DSW_InitOutputBuffer: playCursor = %d, writeCursor = %d\n", playCursor, dsw->dwWriteOffset ); */

  return TRUE;
} // end DSW_InitOutputBuffer()

//---------------------------------------------------------------------------
BOOL DSW_StartOutput( DSoundWrapper *dsw )
{
  HRESULT  hr;
  QueryPerformanceCounter( &dsw->i64LastPlayTime );
  dsw->uLastPlayCursor = 0;
  dsw->dblSamplesPlayedToOutput = 0;
  hr = IDirectSoundBuffer_SetCurrentPosition( dsw->pOutputBuffer, 0 );
  if( hr != DS_OK )
   { DSW_OnOutputError( dsw, "SetCurrentPosition", hr);
     return FALSE;
   }

  // Start the buffer playback in a loop.
  if( dsw->pOutputBuffer != NULL )
   {
     hr = IDirectSoundBuffer_Play( dsw->pOutputBuffer, 0, 0, DSBPLAY_LOOPING );
     if( hr != DS_OK )
      { DSW_OnOutputError( dsw, "Play", hr);
        return FALSE;
      }
     dsw->fOutputOpened = TRUE;
   }

  return TRUE;
} // end DSW_StartOutput()


//---------------------------------------------------------------------------
BOOL DSW_StopOutput( DSoundWrapper *dsw )
{
  HRESULT  hr;

  if( dsw->fOutputOpened ) // <- kludge added 2024-01-06 to reduce multithreading issues
   {  dsw->fOutputOpened = FALSE;  // <- DO NOT CALL ANY AUDIO-OUTPUT-RELATED functions now !
      Sleep(50); // give worker threads and DSW_QueryOutputSpace(), etc, the chance
                 // to take notice, and STOP CALLING any DirectSound-OUTPUT-function.
   }

  // Stop the buffer playback
  if( dsw->pOutputBuffer != NULL )
   {
     hr = IDirectSoundBuffer_Stop( dsw->pOutputBuffer );
     if( hr != DS_OK )
      { DSW_OnOutputError( dsw, "Stop", hr);
        return FALSE;
      }
   }
  return TRUE;
} // end DSW_StopOutput()

//---------------------------------------------------------------------------
void DSW_CloseOutput( DSoundWrapper *dsw )                            // API
{ // Cleanup the sound OUTPUT buffers and the release related resources

  if( dsw->fOutputOpened ) // <- kludge added 2024-01-06 - details in DSW_StopOutput() ...
   {  dsw->fOutputOpened = FALSE;  // <- DO NOT CALL ANY AUDIO-OUTPUT-RELATED functions now !
      Sleep(50); // give worker threads and DSW_QueryOutputSpace(), etc, the chance
                 // to take notice, and STOP CALLING any DirectSound-OUTPUT-function.
   }

  if (dsw->pOutputBuffer)
   {
     IDirectSoundBuffer_Stop( dsw->pOutputBuffer );
     IDirectSoundBuffer_Release( dsw->pOutputBuffer );
     dsw->pOutputBuffer = NULL;
   }
  if (dsw->pDirectSoundOut != NULL )
   {
     IDirectSound_Release( dsw->pDirectSoundOut );
     dsw->pDirectSoundOut = NULL;
   }

} // end DSW_CloseOutput()


//---------------------------------------------------------------------------
BOOL DSW_ZeroEmptySpaceInOutputBuffer( DSoundWrapper *dsw )
{
  HRESULT hr;
  LPBYTE lpbuf1 = NULL;
  LPBYTE lpbuf2 = NULL;
  DWORD dwsize1 = 0;
  DWORD dwsize2 = 0;
  long  bytesEmpty;

  if( ! DSW_QueryOutputSpace( dsw, &bytesEmpty ) ) // updates dblSamplesPlayedToOutput (!)
   { return FALSE; // "last error" (string) already set because the above is an API function
   }
  if( bytesEmpty == 0 )
   { return TRUE;
   }
  // Lock free space in the DS
  hr = IDirectSoundBuffer_Lock( dsw->pOutputBuffer, dsw->dwWriteOffset, bytesEmpty, (void **) &lpbuf1, &dwsize1,
                                  (void **) &lpbuf2, &dwsize2, 0);
  if(hr != DS_OK )
   { DSW_OnOutputError( dsw, "LockOutputBuffer", hr);
     return FALSE;
   }

  // Copy the buffer into the DS
  ZeroMemory(lpbuf1, dwsize1);
  if(lpbuf2 != NULL)
   {
     ZeroMemory(lpbuf2, dwsize2);
   }

  // Update our buffer offset and unlock sound buffer
  dsw->dwWriteOffset = (dsw->dwWriteOffset + dwsize1 + dwsize2) % dsw->iOutputSize;
  IDirectSoundBuffer_Unlock( dsw->pOutputBuffer, lpbuf1, dwsize1, lpbuf2, dwsize2);
  dsw->dblSamplesWrittenToOutput += bytesEmpty / dsw->dwBytesPerFrame_out;
  return TRUE;
} // end DSW_ZeroEmptySpaceInOutputBuffer()

//---------------------------------------------------------------------------
BOOL DSW_QueryOutputSpace( DSoundWrapper *dsw, long *bytesEmpty )      // API
  // Typically called SHORTLY BEFORE 'generating data' for the output,
  //                  and writing those samples via DSW_WriteBlock() .
  // [out] *bytesEmpty : number of BYTES (not *AUDIO SAMPLE POINT*)
  //                  that may be appended to the circular buffer
  //                  via DSW_WriteBlock( with numBytes = *bytesEmpty ) .
  // Return value :
  //   TRUE on succees (even if *bytesEmpty is ZERO = "buffer completely full"),
  //   FALSE if there was a problem, and *bytesEmpty is invalid
  //         (dws->sz255LastOutputError may give a clue about WHAT went wrong).
{
  HRESULT hr;
  DWORD   playCursor;
  DWORD   writeCursor;
  long    numBytesEmpty;
  long    playWriteGap;

  // Protect against "application errors" (avoid crashing with a NULL pointer reference) :
  if( (dsw==NULL) || (dsw->pDirectSoundOut==NULL) ) // POINTERS ARE NOT BOOLEAN !
   { return FALSE;
   }
  if( (dsw->pOutputBuffer==NULL) || ( !dsw->fOutputOpened ) )
   { // ( avoid DSW_QueryOutputSpace() still being called from a worker thread,
     //   when another thread / the "main thread" has already CLOSED or STOPPED the output )
     DSW_SetFirstOutputError( dsw, "QueryOutputSpace called but Output not open" );
     return FALSE; // No valid output buffer ? Refuse to even 'query for output space' !
   }

  // Query to see how much room is in buffer.
  // Note: Even though writeCursor is not used, it must be passed to prevent DirectSound from dieing
  // under WinNT. The Microsoft documentation says we can pass NULL but apparently not.
  // Thanks to Max Rheiner for the fix.
  hr = IDirectSoundBuffer_GetCurrentPosition( dsw->pOutputBuffer, &playCursor, &writeCursor );
  if( hr != DS_OK )
   { DSW_OnOutputError( dsw, "GetCurrentPosition", hr);
     return FALSE;
   }
  // AddTraceMessage("playCursor", playCursor);
  // AddTraceMessage("dwWriteOffset", dsw->dwWriteOffset);

  // Determine size of gap between playIndex and WriteIndex that we cannot write into.
  playWriteGap = writeCursor - playCursor;
  if( playWriteGap < 0 )
   {  playWriteGap += dsw->iOutputSize; // unwrap
   }

  // DirectSound may not have a large enough playCursor so we cannot detect wrap-around.
  // Attempt to detect playCursor wrap-around and correct it.
  if( dsw->fOutputOpened && (dsw->i64CounterTicksPerOutputBuffer.QuadPart != 0) )
   {
     // How much time has elapsed since last check ?
     LARGE_INTEGER   currentTime;
     LARGE_INTEGER   elapsedTime;
     long            bytesPlayed;
     long            bytesExpected;
     long            buffersWrapped;
     QueryPerformanceCounter( &currentTime );
     elapsedTime.QuadPart = currentTime.QuadPart - dsw->i64LastPlayTime.QuadPart;
     dsw->i64LastPlayTime = currentTime;

     // How many bytes does DirectSound say have been played ?
     bytesPlayed = playCursor - dsw->uLastPlayCursor;
     if( bytesPlayed < 0 )
      {  bytesPlayed += dsw->iOutputSize; // unwrap
      }
     dsw->uLastPlayCursor = playCursor;

     // Calculate how many bytes we would have expected to been played by now.
     bytesExpected = (long) ((elapsedTime.QuadPart * dsw->iOutputSize) / dsw->i64CounterTicksPerOutputBuffer.QuadPart);
     buffersWrapped = (bytesExpected - bytesPlayed) / dsw->iOutputSize;
     if( buffersWrapped > 0 )
      {
        // AddTraceMessage("playCursor wrapped! bytesPlayed", bytesPlayed );
        // AddTraceMessage("playCursor wrapped! bytesExpected", bytesExpected );
        playCursor += (buffersWrapped * dsw->iOutputSize);
        bytesPlayed += (buffersWrapped * dsw->iOutputSize);
      }

      // Bookkeeping for the number of sample points actually 'played to the ouput':
      dsw->dblSamplesPlayedToOutput += (bytesPlayed / dsw->dwBytesPerFrame_out);
   }
  numBytesEmpty = playCursor - dsw->dwWriteOffset;
  if( numBytesEmpty < 0 ) numBytesEmpty += dsw->iOutputSize; // unwrap offset

  /* Have we underflowed? */
  if( numBytesEmpty > (dsw->iOutputSize - playWriteGap) )
   {
     if( dsw->fOutputOpened )
      {
        ++dsw->nOutputUnderflows;
        // AddTraceMessage("underflow detected! numBytesEmpty", numBytesEmpty );
      }
     dsw->dwWriteOffset = writeCursor;
     numBytesEmpty = dsw->iOutputSize - playWriteGap;
   }
  *bytesEmpty = numBytesEmpty;
  return TRUE;
} // end DSW_QueryOutputSpace()

//---------------------------------------------------------------------------
BOOL DSW_WriteBlock( DSoundWrapper *dsw, char *buf, long numBytes )    // API
  // Returns TRUE when successful, otherwise FALSE
  //  (and, in the latter case, error-info in dsw->sz255LastOutputError)
  //
{
  HRESULT hr;
  LPBYTE lpbuf1 = NULL;
  LPBYTE lpbuf2 = NULL;
  DWORD dwsize1 = 0;
  DWORD dwsize2 = 0;

  // Protect against "application errors" (avoid crashing with a NULL pointer reference) :
  if( (dsw==NULL) || (dsw->pDirectSoundOut==NULL) ) // POINTERS ARE NOT BOOLEAN !
   { return FALSE;
   }
  if( (dsw->pOutputBuffer==NULL) || ( !dsw->fOutputOpened ) )
   { // ( avoid DSW_QueryOutputSpace() still being called from a worker thread,
     //   when another thread / the "main thread" has already CLOSED or STOPPED the output )
     DSW_SetFirstOutputError( dsw, "WriteBlock called but Output not open" );
     return FALSE; // No valid output buffer ? Refuse to even 'query for output space' !
   }

  // Lock free space in the DS
  hr = IDirectSoundBuffer_Lock( dsw->pOutputBuffer,
         dsw->dwWriteOffset,   // [in] dwWriteCursor :
             // > Offset, in bytes, from the start of the buffer to where the lock begins.
             // > This parameter is ignored if DSBLOCK_FROMWRITECURSOR is specified in the dwFlags parameter.
         numBytes,               // [in] dwWriteBytes :
             // > Size, in bytes, of the portion of the buffer to lock.
             // > Note that the sound buffer is conceptually circular.
         (void **) &lpbuf1,      // [out] lplpvAudioPtr1 :
             // > Address of a pointer to contain the first block of the sound buffer to be locked.
         &dwsize1,               // [out] lpdwAudioBytes1 :
             // > Address of a variable to contain the number of bytes pointed to
             // > by the lplpvAudioPtr1 parameter. If this value is less than the
             // > dwWriteBytes parameter, lplpvAudioPtr2 will point to a second block of sound data.
         (void **) &lpbuf2,      // [out] lplpvAudioPtr2 :
             // > Address of a pointer to contain the second block of the
             // > sound buffer to be locked. If the value of this parameter is NULL,
             // > the lplpvAudioPtr1 parameter points to the entire
             // > locked portion of the sound buffer.
         &dwsize2,               // [out] lpdwAudioBytes2 :
             // > Address of a variable to contain the number of bytes
             // > pointed to by the lplpvAudioPtr2 parameter.
             // > If lplpvAudioPtr2 is NULL, this value will be 0.
         0 ); // [in] dwFlags : Flags modifying the lock event. The following flags are defined:
             // > DSBLOCK_FROMWRITECURSOR :	Locks from the current write position,
             // >   making a call to IDirectSoundBuffer::GetCurrentPosition unnecessary.
             // >   If this flag is specified, the dwWriteCursor parameter is ignored.
             // > DSBLOCK_ENTIREBUFFER : Locks the entire buffer.
             // >   The dwWriteBytes parameter is ignored.
  if(hr != DS_OK )
   { DSW_OnOutputError( dsw, "LockOutputBuffer", hr);
     return FALSE;
   }

  // Copy the buffer into the DS  (possibly "before" and "after" the circular wrapping index)
  CopyMemory(lpbuf1, buf, dwsize1);
  if(lpbuf2 != NULL)
   {
     CopyMemory(lpbuf2, buf+dwsize1, dwsize2);
   }

  // Update our buffer offset and unlock sound buffer
  dsw->dwWriteOffset = (dsw->dwWriteOffset + dwsize1 + dwsize2) % dsw->iOutputSize;
  hr = IDirectSoundBuffer_Unlock( dsw->pOutputBuffer, lpbuf1, dwsize1, lpbuf2, dwsize2);
  dsw->dblSamplesWrittenToOutput += numBytes / dsw->dwBytesPerFrame_out;
  if(hr != DS_OK )
   { DSW_OnOutputError( dsw, "UnlockLockOutputBuffer", hr);
     return FALSE;
   }
  else
   { return TRUE;
   }

} // end DSW_WriteBlock()


//---------------------------------------------------------------------------
DWORD DSW_GetOutputStatus( DSoundWrapper *dsw )
  // Straight from the horse's mouth (besides "DirectSound is a legacy feature"):
  // > This method retrieves the current status of the sound buffer.
  // A-ha. And what does this status tell us; the number of UNUSED BYTES or SAMPLES ?
  //       Of course not. Microsoft's "dwStatus" ..
  // > ... can be a combination of the following flags :
  // >     Flag              Description
  // > DSBSTATUS_BUFFERLOST  The buffer is lost and must be restored before it can be played or locked.
  // > DSBSTATUS_LOOPING     The buffer is being looped. If this value is not set,
  // >                       the buffer will stop when it reaches the end of the sound data.
  // >                       Note that if this value is set, the buffer must also be playing.
  // > DSBSTATUS_PLAYING 	  The buffer is playing. If this value is not set,
  // >                       the buffer is stopped.
{
  DWORD status;

  // Protect against "application errors" (avoid crashing with a NULL pointer reference) :
  if( (dsw==NULL) || (dsw->pDirectSoundOut==NULL) ) // POINTERS ARE NOT BOOLEAN !
   { return DSERR_INVALIDPARAM;
   }
  if( (dsw->pOutputBuffer==NULL) || ( !dsw->fOutputOpened ) )
   { // ( avoid DSW_QueryOutputSpace() still being called from a worker thread,
     //   when another thread / the "main thread" has already CLOSED or STOPPED the output )
     DSW_SetFirstOutputError( dsw, "GetOutputStatus called but Output not open" );
     return DSERR_INVALIDPARAM; // No valid output buffer ? Refuse to even 'query for output status' !
   }

  if (IDirectSoundBuffer_GetStatus( dsw->pOutputBuffer, &status ) != DS_OK)
   { return( DSERR_INVALIDPARAM );
   }
  else
   { return( status );
   }
} // end DSW_GetOutputStatus()



//---------------------------------------------------------------------------
// Audio INPUT ("Capture" a la DirectSound) ... usage summary :
// 1.) Call DSW_OpenInputDevice() with the DESCRIPTIVE NAME (not a "GUID") .
// 2.) Set sample rate, number of channels, and buffer size (thus latency)
//        per DSW_InitInputBuffer(). In Microsoft geek speak, that's
//        IDirectSoundCapture_CreateCaptureBuffer() . Someone (not MS) explained:
//     > The DirectSound capture buffer object, represented by the
//     > IDirectSoundCaptureBuffer8 interface, represents a buffer used for
//     > receiving data from the input device. Like playback buffers,
//     > this buffer is conceptually circular: when input reaches the end
//     > of the buffer, it automatically starts again at the beginning.
//         What follows is from an old "Microsoft DirectX 9.0" SDK, renumbered:
//     > Capturing a sound consists of the following steps:
// 3.) Start the buffer by calling the IDirectSoundCaptureBuffer8::Start method.
//       Normally you should pass DSCBSTART_LOOPING in the dwFlags parameter
//       so that the buffer will keep running continuously rather than stopping
//       when it reaches the end. Audio data from the input device begins filling
//       the buffer from the beginning.
// 4.) Wait until the desired amount of data is available.
//       For one method of determining when the capture cursor reaches a certain point,
//       see Capture Buffer Notification.
// 5.) When sufficient data is available, lock a portion of the capture buffer
//       by calling the IDirectSoundCaptureBuffer8::Lock() method.
//       To make sure you are not attempting to lock a portion of memory
//       that is about to be used for capture, you can first obtain the position
//       of the read cursor by calling IDirectSoundCaptureBuffer8::GetCurrentPosition().
//       As parameters to the Lock method, you pass the size and offset
//       of the block of memory you want to read. The method returns a pointer
//       to the address where the memory block begins, and the size of the block.
//       If the block wraps around from the end of the buffer to the beginning,
//       two pointers are returned, one for each section of the block.
//       The second pointer is NULL if the locked portion of the buffer does not wrap around.
// 6.) Copy the data from the buffer, using the addresses and block sizes
//       returned by the Lock method.
// 7.) Unlock the buffer with the IDirectSoundCaptureBuffer8::Unlock method.
// 8.) Repeat steps 2 to 5 until you are ready to stop capturing data.
//       Then call the IDirectSoundCaptureBuffer8::Stop() method.
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
BOOL DSW_OpenInputDevice(     // aka "Create the capture device object" ...
           DSoundWrapper *dsw,
           char *sz255DevName ) // audio *DEVICE NAME* (not a stuid 'GUID')
  // Return value : TRUE = ok,   FALSE = failure (e.g. device name not found) .
  //     If the function failed, dsw->sz255LastInputError may give a clue why.

{
  GUID *pGUID = NULL;
  HRESULT hr;
  BOOL found_device = FALSE;
  int i;

  strncpy( dsw->sz255InputDeviceName, sz255DevName, 255 );

  if( sz255DevName != NULL ) // use a specific device (not the "Default"-thingy) ?
   { if( dsw->sz255InputDeviceName != sz255DevName ) // <- compare ADDRESS, not STRING CONTENT
      { strncpy( dsw->sz255InputDeviceName, sz255DevName, 255 );
        // '--> this dinosaur will also crash when copying a string to itself !
      }
     for( i=0; (i<dsw->m_nInputDevices) && (i<DSW_MAX_ENUMERATED_DEVICES); ++i)
      { if( strcmp(sz255DevName, dsw->m_sInputDevices[i].sz255DevName) == 0 )
         { if( dsw->m_sInputDevices[i].guid.Data1 != 0 )
            { pGUID = &dsw->m_sInputDevices[i].guid;
            }
           found_device = TRUE;
           break;
         }
      }
   }    // end if( sz255DevName != NULL )
  else
   { dsw->sz255InputDeviceName[0] = '\0';
   }

  if( ! found_device )
   { strcpy( dsw->sz255LastInputError, "OpenInput: Device not found" );
     return FALSE;
   }


  hr = DirectSoundCaptureCreate( pGUID, &dsw->pDirectSoundIn, NULL );
      // ,------------------------------------'                   |
      // '--> "Address of a pointer to a DirectSoundCapture       |
      //       object created in response to this function."      |
      //                                                          |
      //  ,-------------------------------------------------------'
      //  '--> "pUnkOuter: Controlling unknown of the aggregate.
      //                   Its value must be NULL." (the joys of Windows..)
  return ( hr == DS_OK );
} // end DSW_OpenInputDevice()


//---------------------------------------------------------------------------
BOOL DSW_InitInputBuffer(  // aka "Create a Capture Buffer" ...
        DSoundWrapper *dsw,
        unsigned long nFrameRate,
        int nChannels,
        int bytesPerBuffer ) // <- make this AT LEAST long enough for 20 ms of buffering,
        // i.e. nFrameRate * 2 (bytes per 'short') * nChannels / (1000ms/20m),
        // e.g. 48000 Hz   * 2 bytes * 1 channel * (20ms/1000ms) = 1920 sample, "round up" to 2048 samples.
{
  DSCBUFFERDESC  captureDesc;
    // > One of the parameters to the method is a DSCBUFFERDESC structure that
    // > describes the characteristics of the desired buffer.
    // > The last member of this structure is a *pointer to* a WAVEFORMATEX structure,
    // > which must be initialized with the details of the desired WAV format.
    // (a left-over from the ancient Windows "MultiMedia" system)
  WAVEFORMATEX   wfFormat;
  HRESULT        hr;

  if( (dsw==NULL) || (dsw->pDirectSoundIn==NULL) )
   { return FALSE; // "application error" (avoid crashing with a NULL pointer reference)
   }

  // Define the buffer format  (in the DirectX SDK: "Set up WAV format structure", e.g. in CreateBasicBuffer)
  memset( &wfFormat, 0, sizeof(WAVEFORMATEX) );
  wfFormat.wFormatTag      = WAVE_FORMAT_PCM;
  wfFormat.nChannels       = nChannels;
  wfFormat.nSamplesPerSec  = nFrameRate;
  wfFormat.wBitsPerSample  = 8 * sizeof(short);
  wfFormat.nBlockAlign     = wfFormat.nChannels * (wfFormat.wBitsPerSample / 8);
  wfFormat.nAvgBytesPerSec = wfFormat.nSamplesPerSec * wfFormat.nBlockAlign;
  wfFormat.cbSize          = 0;   /* No extended format info. */
  dsw->dwInputFifoSizeInBytes   = bytesPerBuffer;
  dsw->dwBytesPerFrame_in  = nChannels * sizeof(short);


  // ----------------------------------------------------------------------
  // Setup the secondary buffer description
  ZeroMemory(&captureDesc, sizeof(DSCBUFFERDESC));
  captureDesc.dwSize = sizeof(DSCBUFFERDESC);
  captureDesc.dwFlags =  0;
  captureDesc.dwBufferBytes = bytesPerBuffer;
  captureDesc.lpwfxFormat = &wfFormat;

  // Create the capture buffer
  if ((hr = IDirectSoundCapture_CreateCaptureBuffer( dsw->pDirectSoundIn,
                &captureDesc, &dsw->pInputBuffer, NULL)) != DS_OK)
   { DSW_OnInputError( dsw, "CreateCaptureBuffer", hr );
     return FALSE;
   }
  dsw->uReadOffset = 0;  // reset last read position to start of buffer
       // (somehow similar with a FIFO tail index in a classic lock-free circular FIFO.
       //  The FIFO head index is SIMILAR TO what Microsoft calls "capture cursor";
       //  see text quoted from a stoneage 'Direct-something SDK' further above)

  // Prepare our 'helpers for bookkeeping' / detect timing problems, overflows, etc
  dsw->nInputOverflows         = 0;
  dsw->dblInputSamplingRate_Hz = (double)nFrameRate;
  dsw->dblInputWrapInterval_s  = (double)bytesPerBuffer / (double)( sizeof(short) * nFrameRate * nChannels );
       // '--> if the recommended buffer size was specified,
       //      the result should be at least 0.02 seconds !
  dsw->dblLastInputTime_s = dsw->dblTimeAtLastCaptureHistoryEntry_s = DSW_ReadHighResTimestamp_s();


  // Do NOT set dsw->fInputOpened here yet, because without also calling
  // DSW_StartInput(), some mimosa deep inside DirectSound crashed
  //     when a worker thread tried to call e.g. DSW_ReadBlock() !


  return TRUE;
} // end DSW_InitInputBuffer()


//---------------------------------------------------------------------------
BOOL DSW_StartInput( DSoundWrapper *dsw )
{
  HRESULT hr;

  if( (dsw==NULL) || (dsw->pDirectSoundIn==NULL) )
   { return FALSE; // "application error" (avoid crashing with a NULL pointer reference)
   }

  dsw->dwCaptureHistoryIndex = 0;


  // Start the capturing into the input buffer
  if( dsw->pInputBuffer != NULL )
   {
     hr = IDirectSoundCaptureBuffer_Start( dsw->pInputBuffer, DSCBSTART_LOOPING );
     dsw->dblLastInputTime_s = DSW_ReadHighResTimestamp_s();
     if( hr != DS_OK )
      { DSW_OnInputError( dsw, "StartInput", hr );
        return FALSE;
      }
     dsw->fInputOpened = TRUE;  // allow calling e.g.
          // IDirectSoundCaptureBuffer_GetCurrentPosition(),
          // IDirectSoundCaptureBuffer_Lock(),
          // IDirectSoundCaptureBuffer_Unlock(),
          // and who-knows-what may be called from a 'worker thread'.

     return TRUE;
   }
  return FALSE;  // nothing to START
} // end DSW_StartInput()


//---------------------------------------------------------------------------
BOOL DSW_StopInput( DSoundWrapper *dsw )
{
  HRESULT hr;

  if( dsw->fInputOpened ) // <- kludge added 2024-01-06 to reduce multithreading issues
   {  dsw->fInputOpened = FALSE;  // <- DO NOT CALL ANY AUDIO-INPUT-RELATED functions now !
      Sleep(50); // give some worker thread the chance to stop calling us
                 // before we REALLY "stop the input"; and prevent any of our
                 // 'wrapper functions' to invoke INPUT-related DirectSound functions.
   }

  if( (dsw==NULL) || (dsw->pDirectSoundIn==NULL) )
   { return FALSE; // "application error" (avoid crashing with a NULL pointer reference)
   }

  // Stop capturing" (feeding samples into the input buffer)
  if( dsw->pInputBuffer != NULL )
   {
     hr = IDirectSoundCaptureBuffer_Stop( dsw->pInputBuffer );
     if( hr != DS_OK )
      { DSW_OnInputError( dsw, "StopInput", hr );
        return FALSE;
      }
     return TRUE;
   }
  return FALSE;  // nothing to STOP
} // end DSW_StopInput()


//---------------------------------------------------------------------------
BOOL DSW_QueryInputFilled( DSoundWrapper *dsw, long *bytesFilled )
  // This function must be sufficiently robust to NOT CRASH
  //      when called even if Mr. DirectSound's *input* is not open at all !
  //      For reliable 'bookkeeping' (watch for possible input overruns, etc),
  //      the time between calls of DSW_QueryInputFilled() should not exceed
  //      HALF the buffer time (=time it takes to fill the entire buffer
  //      allocated by DSW_InitInputBuffer() .. e.g. at least every 10 ms
  //      when DirectSound's input buffer ("capture buffer") can store 20 ms.
{
  HRESULT hr;
  DWORD dwCapturePos;
  DWORD dwReadPos;
  long  filled;
  double dblCurrentTime_s, dblSecondsSinceLastCall;
  int    nBytesCapturedSinceLastCall;
  T_DSW_InputCaptureHistoryEntry *pCH;


  dblCurrentTime_s = DSW_ReadHighResTimestamp_s();


  // Protect against "application errors" (avoid crashing with a NULL pointer reference) :
  if( (dsw==NULL) || (dsw->pDirectSoundIn==NULL) )
   { return FALSE;
   }
  if( dsw->pInputBuffer==NULL )
   { DSW_SetFirstOutputError( dsw, "DSW_QueryInputFilled called but Input not open" );
     return FALSE; // No valid input buffer ? Refuse to even 'query for input' !
   }

  // Query to see how much data is in buffer.
  // We don't need the capture position but sometimes DirectSound
  // doesn't handle NULLS correctly
  // so let's pass a pointer just to be safe.
  hr = IDirectSoundCaptureBuffer_GetCurrentPosition( dsw->pInputBuffer, &dwCapturePos, &dwReadPos );
  // ,----------'
  // '--> From a MORE RECENT description (about "IDirectSoundCaptureBuffer8", note the EIGHT) :
  // > Capture Buffer Notification
  // > To copy data out of the capture buffer at regular intervals,
  // > your application has to know when the read cursor, which is the point
  // > up to which it is safe to read data, reaches certain points in the buffer.
  // > One way to get this information is to poll the position of the read cursor
  // > using IDirectSoundCaptureBuffer8::GetCurrentPosition.   [WB: .. as used here..]
  // >
  // > A more efficient way is to use notifications.
  // > Using the IDirectSoundNotify8::SetNotificationPositions() method,
  // > you can set any number of points within the buffer where events
  // > are to be signaled. You cannot do this while the buffer is running. (..)
  // >
  if( hr != DS_OK ) // IDirectSoundCaptureBuffer_GetCurrentPosition() didn't co-operate ->
   { DSW_OnInputError( dsw, "GetCurrentInputPos", hr );
     return FALSE;
   }

  //           __________________________________________________________   .
  //          |    Circular "Capture Buffer"                             |  .
  //          |             |<-------- "still"  VALID------------>|      |  .
  //          |__________________________________________________________|  .
  //                       /|\             /|\                   /|\        .
  //                        |               |                     |         .
  //               dwCapturePos:            |                     |         .
  //               ~in use by the           |<------"filled"----->|         .
  //                audio driver ?      uReadOffset               |         .
  //                                    ~~ our FIFO tail index   dwReadPos  .
  //
  // > "The capture cursor is ahead of the read cursor. (???)
  // >  The data after the read position up to
  // >  and including the capture position
  // >  is not necessarily valid data."      (??? .. How does this line up with the above ???)
  //     Hmmm.  Why did one of the first calls of IDirectSoundCaptureBuffer_GetCurrentPosition()
  //            return "dwCapturePos" = 1920,  but "dwReadPos" = 0 then ?
  //            Sounds more like "the data after dwReadPos, up to CapturePos"
  //            CONTAIN VALID DATA that we may LOCK / READ / UNLOCK ?
  if(  (dwCapturePos != dsw->dwPrevInputCapturePos)
    || (dwReadPos    != dsw->dwPrevInputReadPos) )
   { // "something happened" (since the last call of IDirectSoundCaptureBuffer_GetCurrentPosition)
     // so emit this to the "capture history index". Maybe this allows us
     // to understand what this funny-named "Capture"- versus "Read"-Position is about,
     // and why > "The data after the read position up to and including the
     //         >  capture position is not necessarily valid data."
     pCH = &dsw->sCaptureHistory[ dsw->dwCaptureHistoryIndex ];
     dsw->dwCaptureHistoryIndex = ( dsw->dwCaptureHistoryIndex+1 ) & 255;
     nBytesCapturedSinceLastCall = (int)dwCapturePos - (int)dsw->dwPrevInputCapturePos;
     if( nBytesCapturedSinceLastCall < 0 )
      {  nBytesCapturedSinceLastCall += dsw->dwInputFifoSizeInBytes; // circular FIFO wrapped around
      }
     dsw->dblSamplesCapturedFromInput += nBytesCapturedSinceLastCall / dsw->dwBytesPerFrame_in;

     pCH->dwCapturePos = dsw->dwPrevInputCapturePos = dwCapturePos;
     pCH->dwReadPos    = dsw->dwPrevInputReadPos    = dwReadPos;
     pCH->dwTotalSamplesRead = dsw->dwTotalSamplesReadFromInput;
     pCH->iDeltaT_ms   = (int)( 0.5 + 1e3 * (dblCurrentTime_s - dsw->dblTimeAtLastCaptureHistoryEntry_s) );
     dsw->dblTimeAtLastCaptureHistoryEntry_s = dblCurrentTime_s;
     if( dsw->dwCaptureHistoryIndex == 255 )
      {  dsw->dwCaptureHistoryIndex = dsw->dwCaptureHistoryIndex; // <- place for a "conditional breakpoint" !
      }
     // WATCHED the result with Borland's debugger ( "MyDirectSound.sCaptureHistory[0] .. [10]" ),
     //   when using CW_DSP_INPUT_BUFFER_NSAMPLES = 8192 = circa 170 milliseconds "bufferable"...
     //  ->     dwCapturePos  dwReadPos  TotalSamplesRead  delta t in ms
     //    [0] = {   2880,         0,        0        ,   40 }
     //    [1] = {   4800,      3000,        0        ,   20 }
     //    [2] = {   5760,      6000,      1500       ,   10 }
     //    [3] = {   6720,      6000,      1500       ,   8  }
     //    [4] = {   7680,      6000,      2524       ,   11 } // <- application's input buffer was full, it drained the buffer at max 2048 bytes per 'Read()'
     //    [5] = {   8640,      6000,      3000       ,   12 }
     //    [6] = {   9600,      9000,      3000       ,   6  }
     //    [7] = {  10560,      9000,      4500       ,   11 }
     //    [8] = {  11520,      9000,      4500       ,   11 }
     //    [9] = {  12480,     12000,       ...       ,   12 }
     //   [10] = {  13440,     12000,                 ,   6  }
     //               '--> ok, confirms "dwCapturePos" *LEADS* "dwReadPos",
     //                    and that "dwReadPos" doesn't have anything to do
     //                    with the application's own BUFFER TAIL INDEX !
     //     So "dwCapturePos" is THE REAL BUFFER HEAD INDEX ("occupied by the hardware").
     //     But why the heck does "dwReadPos" always step by 3000, if it steps at all ? ?
     //     At fsample=48 kHz, 3000 samples = 62.5 bloody milliseconds ! ! ? !
     //     As a test, using the 'Remote CW Keyer' application,
     //     the almost-latency-free SIDETONE FROM THE COM-PORT'S TXD OUTPUT
     //     was 'echoed back' into the PC's onboard microphone, and plotted
     //     along with the original keying signal in the program's timing scope.
     //     Result: Total latency of originally 138 ms - see screenshot in
     //        Remote_CW_Keyer/manual/MIC_input_latency_with_DirectSound.png !
     //
   } // end if < "capure pos" or "read pos" have changed since the last call >

  filled = dwReadPos - dsw->uReadOffset;
  if( filled < 0 )
   {  filled += dsw->dwInputFifoSizeInBytes; // unwrap offset
   }
  *bytesFilled = filled;

  // Do some "bookkeeping" for DSoundWrapper.nInputOverflows :
  dblSecondsSinceLastCall = dsw->dblLastInputTime_s - dblCurrentTime_s;
  if( dblSecondsSinceLastCall >= dsw->dblInputWrapInterval_s )
   { // Calling interval is too slow (by a factor of two) ..
     ++dsw->nInputOverflows;
   }

  return TRUE;
} // end DSW_QueryInputFilled()


//---------------------------------------------------------------------------
BOOL DSW_ReadBlock( DSoundWrapper *dsw, char *buf, long numBytes )
{
  HRESULT hr;   // what does the "H" stand for, Microsoft ? Is this a HANDLE ? ?
  LPBYTE lpbuf1 = NULL;
  LPBYTE lpbuf2 = NULL;
  DWORD dwsize1 = 0;
  DWORD dwsize2 = 0;
  int nSamplePoints;
#if ( SWI_AUDIO_INPUT_DUMMY )
  int   i, iSample;
#endif // SWI_AUDIO_INPUT_DUMMY ?


  HERE_I_AM__DSW(); // -> DSW_iLastSourceLine = __LINE__ (when compiled for "hardcore debugging")

  if( (dsw==NULL) || (dsw->pDirectSoundIn==NULL) )
   { HERE_I_AM__DSW();
     DSW_SetFirstInputError( dsw, "Called DSW_ReadBlock() but there's no DSOUND instance." );
     return FALSE; // "application error" (avoid crashing with a NULL pointer reference)
   }

  if( dsw->pInputBuffer==NULL )
   { HERE_I_AM__DSW();
     DSW_SetFirstInputError( dsw, "Called DSW_ReadBlock() but there's no INPUT BUFFER." );
     return FALSE; // "application error" (avoid crashing in ..GetCurrentPosition() )
   }

  if( (!dsw->fInputOpened ) // oops.. in the meantime, someone has called DSW_StopInput() ?!
    || (dsw->dwBytesPerFrame_in == 0 ) ) // on this occasion, also avoid crashing from div-by-zero
   { // (this may indicate a multi-threading issue, so beware..)
     HERE_I_AM__DSW();
     DSW_SetFirstInputError( dsw, "Called DSW_ReadBlock() but input isn't open." );
     return FALSE;
   }

  // Refuse to read a number of bytes that is NOT an integer multiple
  //  of a "frame" (aka "sample point, with N channels") :
  nSamplePoints = numBytes / dsw->dwBytesPerFrame_in;
  if( numBytes != ( nSamplePoints * dsw->dwBytesPerFrame_in ) )
   { HERE_I_AM__DSW();
     DSW_SetFirstInputError( dsw, "Called DSW_ReadBlock() with a broken blocksize." );
     return FALSE;
   }


  // Lock free space in the DS
  HERE_I_AM__DSW();  // <- 2024-01-06 : Last sign of life HERE, before crashing deep inside in IDirectSoundCaptureBuffer_Lock() !
  hr = IDirectSoundCaptureBuffer_Lock( dsw->pInputBuffer,
             dsw->uReadOffset, // [in] DWORD dwReadCursor :
                                  //      Offset, in bytes, from the start of the buffer to where the lock begins.
             numBytes,            // [in] DWORD dwReadBytes  :
                                  //      Size, in bytes, of the portion of the buffer to lock.
                                  //      Note that the capture buffer is conceptually circular.
             (void **) &lpbuf1,   // [out] LPVOID* lplpvAudioPtr1  : Address of a pointer to contain
                                  //       the first block of the capture buffer to be locked
             &dwsize1,            // [out] LPDWORD lpdwAudioBytes1 : Address of a variable to contain
                                  //       the number of bytes pointed to by the lplpvAudioPtr1 parameter.
                                  //       If this value is less than the dwReadBytes parameter,
                                  //       lplpvAudioPtr2 will point to a second block of data.
             (void **) &lpbuf2,   // [out] LPVOID* lplpvAudioPtr2  : Address of a pointer to contain
                                  //       the second block of the capture buffer to be locked.
                                  //       If the value of this parameter is NULL, the lplpvAudioPtr1
                                  //       parameter points to the entire locked portion of the capture buffer.
             &dwsize2,            // [out] LPDWORD lpdwAudioBytes2 : Address of a variable to contain
                                  //       the number of bytes pointed to by the lplpvAudioPtr2 parameter.
                                  //       If lplpvAudioPtr2 is NULL, this value will be 0.
             0 );                 // [in] dwFlags : 0 or DSCBLOCK_ENTIREBUFFER (to lock the ENTIRE buffer).
  HERE_I_AM__DSW();
  if (hr == DS_OK)
   {
     // Copy from DS to the buffer
     HERE_I_AM__DSW();
     CopyMemory( buf, lpbuf1, dwsize1);
     HERE_I_AM__DSW();
     if(lpbuf2 != NULL)
      { HERE_I_AM__DSW();
        CopyMemory( buf+dwsize1, lpbuf2, dwsize2);
        HERE_I_AM__DSW();
      }

     // Update our buffer offset and unlock sound buffer
     dsw->uReadOffset = (dsw->uReadOffset + dwsize1 + dwsize2) % dsw->dwInputFifoSizeInBytes;
     HERE_I_AM__DSW();  // 2024-01-06 : IDirectSoundCaptureBuffer_Unlock()
                        //  (called a few lines below, after the 'debug stuff')
                        // didn't return, causing problems in CwDSP.c : CwDSP_Stop()
                        // when trying to terminate the calling thread
                        // of DSW_ReadBlock() "politely" .
     nSamplePoints = ( dwsize1 + dwsize2 ) / dsw->dwBytesPerFrame_in;
#  if( SWI_HARDCORE_DEBUGGING )
     // Take a snapshot of the parameters passed to IDirectSoundCaptureBuffer_Unlock() below...
     dsw->dwDebug[0] = __LINE__;
     dsw->dwDebug[1] = (DWORD)dsw->pInputBuffer;
     dsw->dwDebug[2] = (DWORD)lpbuf1;
     dsw->dwDebug[3] = (DWORD)dwsize1;
     dsw->dwDebug[4] = (DWORD)lpbuf2;
     dsw->dwDebug[5] = (DWORD)dwsize2;
     //  WATCHED the following global varibles after getting stuck in CwDSP_Stop()
     //  because DSW_ReadBlock() -> IDirectSoundCaptureBuffer_Unlock() didn't return:
     //    MyDirectSound.fInputOpened    = 1 (ok)
     //    MyDirectSound.fOutputOpened   = 1 (ok)
     //    MyDirectSound.pInputBuffer = 0x00C3D8E0 (ok)
     //    MyDirectSound.uReadOffset = 1536       (ok)
     //    MyDirectSound.dwInputFifoSizeInBytes  = 4096
     //    MyDirectSound.dwDebug[0]  = 1227 = __LINE__ (above,ok)
     //    MyDirectSound.dwDebug[1]  = 0x00C3D8E0 = dsw->pInputBuffer (ok)
     //    MyDirectSound.dwDebug[2]  = 0x00C3CC30 = lpbuf1
     //    MyDirectSound.dwDebug[3]  = 0x00000200 = dwsize1
     //    MyDirectSound.dwDebug[4]  = 0x00000000 = lpbuf2
     //    MyDirectSound.dwDebug[5]  = 0x00000000 = dwsize2
#  endif // SWI_HARDCORE_DEBUGGING ?
     IDirectSoundCaptureBuffer_Unlock( dsw->pInputBuffer, lpbuf1, dwsize1, lpbuf2, dwsize2);
     HERE_I_AM__DSW();   // survived the call of IDirectSoundCaptureBuffer_Unlock()
#   if(SWI_AUDIO_INPUT_DUMMY) // (1)=TEST with an "artifical input signal" (sawtooth), (0)=normal compilation
     for( i=0; i<nSamplePoints; ++i )
      { iSample = (int)( (dsw->dwTotalSamplesReadFromInput % dsw->dwInputFifoSizeInBytes )
                            * 65535 / dsw->dwInputFifoSizeInBytes ) - 32767;
        buf[ i*dsw->dwBytesPerFrame_in + 0 ] = iSample & 0x00FF; // lower 8 bits
        buf[ i*dsw->dwBytesPerFrame_in + 1 ] = iSample >> 8;     // upper 8 bits
        ++dsw->dwTotalSamplesReadFromInput;
      }
#   else // no TEST but a NORMAL COMPILATION :
     dsw->dwTotalSamplesReadFromInput += nSamplePoints; // <- overflows after a few hours
#   endif  // SWI_AUDIO_INPUT_DUMMY ?
     dsw->dblSamplesReadFromInput += nSamplePoints; // <- won't overflow during the lifetime of any computer
     HERE_I_AM__DSW();
     return TRUE;
   }
  else // IDirectSoundCaptureBuffer_Lock() refused to lock the wanted number of BYTES, beginning at our "uReadOffset" ...
   { HERE_I_AM__DSW();
     return FALSE;
   }
} // end DSW_ReadBlock()

//---------------------------------------------------------------------------
void DSW_CloseInput( DSoundWrapper *dsw )                            // API
{ // Cleanup the sound INPUT buffers and the release related resources

  if( dsw->fInputOpened ) // <- similar as in DSP_StopInput() .. avoid multithreading issues
   {  dsw->fInputOpened = FALSE;  // <- DO NOT CALL ANY AUDIO-INPUT-RELATED functions now !
      Sleep(50); // wait until 'worker threads' have taken notice,
                 // and e.g. DSW_ReadBlock() refuses to use pInputBuffer,
                 // because (in contrast to its name), pInputBuffer
                 // isn't just the circular input buffer, but also contains
                 // the pointers to the 'IDirectSound'-object's METHOD TABLE !
                 // See definition of e.g. IDirectSoundCaptureBuffer_Lock()
                 // in Microsoft's "dsound.h" :
                 // #define IDirectSoundCaptureBuffer_Lock(p,a,b,c,d,e,f,g)  (p)->lpVtbl->Lock(p,a,b,c,d,e,f,g)
                 //                                        '--------->--------'
                 // The 'p' in the definition above is our
                 // "LPDIRECTSOUNDCAPTUREBUFFER pInputBuffer"
                 //    |______________________|
   } // end if( dsw->fInputOpened )



  if( dsw->pInputBuffer != NULL ) // <- this is a pointer, not a BOOLEAN
   {
     IDirectSoundCaptureBuffer_Stop( dsw->pInputBuffer );
     IDirectSoundCaptureBuffer_Release( dsw->pInputBuffer );
     dsw->pInputBuffer = NULL;
   }
  if (dsw->pDirectSoundIn != NULL )
   {
     IDirectSoundCapture_Release( dsw->pDirectSoundIn );
     dsw->pDirectSoundIn = NULL;
   }
} // end DSW_CloseInput()


//---------------------------------------------------------------------------
void DSW_OnOutputError( DSoundWrapper *dsw, char *pszInfo, HRESULT hr )
  // Invoked from a couple of OUTPUT functions after problems
  // with a DSound function call... good place for a breakpoint !
{
  sprintf( dsw->sz255LastOutputError, "%s:%d", pszInfo, (int)hr );
}

//---------------------------------------------------------------------------
void DSW_SetFirstOutputError( DSoundWrapper *dsw, char *pszInfo )
{ if( dsw->sz255LastOutputError[0] == '\0' ) // only if this the FIRST error..
   { strcpy( dsw->sz255LastOutputError, pszInfo );
   }
}

//---------------------------------------------------------------------------
void DSW_OnInputError( DSoundWrapper *dsw, char *pszInfo, HRESULT hr )
  // Invoked from a couple of INPUT functions after problems
  // with a DSound function call... good place for a breakpoint !
{
  sprintf( dsw->sz255LastInputError, "%s:%d", pszInfo, (int)hr );
}

//---------------------------------------------------------------------------
void DSW_SetFirstInputError( DSoundWrapper *dsw, char *pszInfo )
{ if( dsw->sz255LastInputError[0] == '\0' ) // only if this the FIRST error..
   { strcpy( dsw->sz255LastInputError, pszInfo );
   }
}

